#!/usr/bin/env python2.6

"""
picbl.py   copyright 2006 Frank Sergeant (frank@pygmy.utoh.org)
           MIT/BSD style license,
           (http://pygmy.utoh.org if you wish to see the exact license)

  Version 20060722

  Download an "Intel hex" file of a PIC program, via a serial port,
  into a Microchip PIC microprocessor which is running my bootloader.

  Many bootloader approaches are possible.  This particular combination
  tries to put as much logic on the PC side (where memory is plentiful)
  and the least on the PIC side (where memory is limited).  Also, it
  uses serial transmit on the PC and a bit-banged serial receive on
  the PIC, but no serial transmit from the PIC back to the PC.  The
  PIC's serial input pin can be any I/O pin or, probably best, the
  input-only RA5/Vpp/MCLR* pin.  If RA5 is used, the voltage on that
  line should be scaled or clamped so that it does not get anywhere
  near the nominal 13 V that would put the PIC into high-voltage
  programming mode.

  Picbl is set up for a 5 ms delay at appropriate points, which should
  be plenty of time for the 16F818 or 16F819 and probably enough for
  the 16F877 (and other 16F87x chips).  Picbl delays right after sending
  the 2 bytes that begin a 32-word block (so the 16F819 has time to do
  the block erase) and also after sending every fourth word (so the
  16F819 has time to burn the four words in its buffer into the actual
  flash memory)

  For other PICs, the delay value or positions might need to be changed.

  The 16F819 and the 16F877 program the flash in blocks of four words
  (8 bytes) and the actual programming does not occur until the last
  word of the four-word block has been sent.  So, Picbl must send
  multiples four words.  If the hex file should end otherwise then
  Picbl will pad out the final four-word block with 0xFF bytes.

  The checksums in the Intel hex file *are* checked and an error is
  reported and the program halts if the checksum is bad.  If you wish
  to hand edit the hex file and not fix up the checksums, then you
  could comment out the checksum tests in this file (picbl.py).

  From the Intel hex file, Picbl generates the entire binary image
  in memory prior to beginning to transmit to the PIC.  The first
  32 words starting at address 0x0000 are *skipped*.  The first data
  to be transmitted to the PIC correspond to the image at word address
  32 (i.e. 0x0020), byte address 64 (i.e. 0x0040).  This is because
  the bootloader does not wish to erase and overwrite the first 32-word
  block which includes the bootloader entry point, interrupt vector, and
  jump to the main bootloader code in high memory.  So, the bootloader
  starts programming the flash at the second block (word address 0x0020).
  
  To use this program, start picbl.py first, then power up the PIC
  chip within a few seconds.  Picbl sends a synchronizing character
  for the first few seconds, then sends the count of words, then sends
  the binary image.

  For this particular version of Picbl and the corresponding bootloader,
  the binary image is transmitted blind, with no feedback at all from
  the PIC.  Picbl always uses serial parameters of 9600,n,8,1 with no
  hardware and no software flow control.  The serial port is specified
  as a zero-based number.  So

        under DOS/Windows,
           port 0 would correspond to COM1:
           port 1 would correspond to COM2:

        under Linux,
           port 0 would correspond to /dev/ttyS0
           port 1 would correspond to /dev/ttyS1

  The Intel hex file is expected to be the "inxh32" style
  file produced by the gpasm PIC assembler, but probably
  any Intel hex file will work.
  
  Usage:
            picbl <intelhexfile> <serialport>
            
  e.g. to send the dummy1.hex file to the PIC via
  the first serial port (known typically as
  either COM1: or as /dev/ttyS0)
           
            picbl dummy1.hex 0

  Intel hex file notes:

  Here is a snippet from a dummy hex file (with a number of
  lines in the middle omitted for brevity:

--------- dummy1.hex ----------------------------
:020000040000FA
:020000002428B2
:040008008A01202821
:1000400009000000000000003A200D30A7200A300F
:10005000A7204430A7205530A7204D30A7204D3091
:10006000A7205930A7203130A7207F207F207F2074
 ....
:1001E000413442344334443445344634AC003C302A
:1001F000A7202C0EE020A7202C08E020A7203E30CE
:04020000A72008002B
:02400E00103F61
:00000001FF
--------------------------------------------------

and here is the same snippet with extra spacing around fields
and with comments in parentheses following each line:

(   len addr type data                                 cksum)

   : 02 0000 04   0000                                    FA
( A type 4 record's data represents bits 31..16 of a 32-bit
  address.  We do not need this record for the PICs this
  is designed to work with.  However, if it is present,
  the data must be '0000' as above)
   
   : 02 0000 00   2428                                    B2
( A type 0 record represents the binary image to be transmitted to
  the PIC and the address.  In this case, the data consists of two
  bytes that go at absolute address '0000'.  The address is in
  big-endian format, with 2 ascii character hex digits representing one
  byte.  The data '2428' represents one 16-bit word (two bytes) in
  little-endian format.  So, the byte that should go at the very
  first location in the PIC's memory is '28' and the next byte (at
  byte address 0001 in the image is '24'.  The listing produced
  by the assembler would show the value as '2824',   This represents
  the instruction 'goto UserApp' where UserApp's word address is
  0x0024.  If you look at the PIC16F819 instruction set documentation,
  you will see the goto instruction as '10 1kkk kkkk kkkk' which
  represents 14 bits.  We (and the hex file, etc) pad the 14 bits with
  two zero bits on the left (so we can deal in multiples of an 8-bit
  byte).  This gives '0010 1kkk kkkk kkkk'.  The 11 bits shown as 'k'
  are the destination address of the goto instruction.  In this example
  the address should be the 11 least significant bits of 0x0024 (i.e.
  the word address of the UserApp label) which are '00000100100' in
  binary.  Thus, after plugging in the adrress in place of the 'k's, we
  get '0010 1000 0010 0100' which is 0x2824 in hexadecimal.  So, the
  word value 0x2824 is shown in the intel hex type 2 record as a value
  of '2428' (i.e. little-endian format with the least significant byte
  first and the most significant byte second).  The checksum 'B2' is
  chosen so that if all the bytes values are added together, the least
  significant 8 bits will be zero.  E.g.
   02 + 00 + 00 + 00 + 0x24 + 0x28 + 0xb2  =  0x100
  )
   
   : 04 0008 00   8A012028                                21
( This is another type 0 record.  The byte address is '0008' which
  is byte address 0x0008, which corresponds to a word address of 0x0004.
  Note that there is a gap between the previous line which had 2 bytes
  at byte address 0x0000 and this line which starts at byte address 0x0008.
  Picbl would fill in the gap with six pad bytes of 0xFF.  So, the value
  of the binary image at byte address 0x0008 is shown as '8A012028' which
  represents 2 16-bit words (4 bytes).  Since it is in little-endian format,
  the byte at byte address 0x0008 is 0x01 and the byte at byte address 0x0009
  is 0x8A, etc.  A hex dump would show that part of the image as
            addr  
            0008  01 8a 28 20
)
   
   : 10 0040 00   09000000000000003A200D30A7200A30        0F
( This is another type 0 record for the 16 (0x10) bytes, or 8 words
  at byte address 0x0040 or word address 0x0020.  Note that whereas
  the data '09000000....0A30' is in 16-bit pairs of little-endian
  bytes while the address '0040' is in straight big-endian format.
)
   
   : 10 0050 00   A7204430A7205530A7204D30A7204D30        91
   : 10 0060 00   A7205930A7203130A7207F207F207F20        74
 ....
   : 10 01E0 00   413442344334443445344634AC003C30        2A
   : 10 01F0 00   A7202C0EE020A7202C08E020A7203E30        CE
   
   : 04 0200 00   A7200800                                2B
( This type 0 record happens to be the last line of data for our
  sample program.  It represents the bytes 0x20 0xA7 0x00 0x08
  at byte address 0x0200 (word address 0x0100).)
   
   : 02 400E 00   103F                                    61
( This type 0 record represents the configuration word 0x3F10 to
  be burned into word address 0x2007 (which is byte address 0x400E)
  when programming the PIC with a hardware programmer rather than
  with a bootloader.  This record is not required by the Picbl and
  is ignored if it is present.  That way, the same application source
  code, and even the same Intel hex file, can be used either with Picbl
  and the bootloader or with a hardware programmer and no bootloader.)
   
   : 00 0000 01                                           FF
( This type 1 record is simply an end-of-file marker.  Note the bytes
  still add up to a number whose least significant 8 bits are zero
  00 + 00 + 00 + 01 + 0xFF  =  0x100
)


"""


# Communicate with PIC bootloader

import sys
import serial
import time
import threading

SYNC = 0xA5           # must be the same used in boot.asm

BLOCKDELAY = 0.005    # 5 milliseconds -- should work for '818, '819, '87x
BYTEDELAY =  0.005    # we might reduce this to 0.002 or 0.003


def abortWithUsageMessage():
    msg = """
Usage:
      picbl <intelhexfile> <serialportnumber>
  e.g. to download the Intel hex file named 'dummy1.hex' via the
  first serial port (COM1: under DOS/Windows, /dev/ttyS0 under Linux)
  say:
      picbl dummy1.hex 0
"""
    print msg
    sys.exit(1)

def parseCommandLine():
    """Validate the command line and open the hex file and the serial port,
       setting global variables to refer to them.
    """
    global portNumber, hexFile, serialPort
    if len(sys.argv) != 3:
        print "Wrong number of command line arguments"
        abortWithUsageMessage()

    try:
        portNumber = sys.argv[2]
    except:
        print "The port number %s should be an integer between 0 and 7" % sys.argv[2]
        abortWithUsageMessage()

#    if not (0 <= portNumber < 8): 
#        print "The port number %s should be an integer between 0 and 7" % sys.argv[2]
#        abortWithUsageMessage()

    try:
        #hexFile = open (sys.argv[1], "r")
        hexFile = open (sys.argv[1], "rU")
    except:
        print "Cannot open hex file named %s" % sys.argv[1]
        abortWithUsageMessage()

    try:
        serialPort = openSerialPort(portNumber)
    except:
        print "Cannot open serial port number %d" % portNumber
        abortWithUsageMessage()
        
    print "Okey dokey, we'll send the file %s to serial port %s" % (sys.argv[1], portNumber)

def openSerialPort(num):
    """Open the /dev/ttyS0 or /dev/ttyS1 etc, specifying everything
       even though most parameters default to what we want.
       'timeout=0' means non-blocking
    """
    ser = serial.Serial(port=num,
                        baudrate=4800,
                        bytesize=serial.EIGHTBITS,
                        parity=serial.PARITY_NONE,
                        stopbits=serial.STOPBITS_TWO,
                        timeout=0, xonxoff=0, rtscts=0,
                        writeTimeout=None)
    return ser



class hexrec(object):
    """Just a simple object to attach a few fields to"""
    def __init__(self, record):
        self.record = record.strip()
        self.goodCheckSum = self.checkCheckSum()
        if not self.goodCheckSum:
            print "bad checksum on record:"
            print "'%s'" % self.record
            abortWithUsageMessage()
        self.parseRecord()

    def checkCheckSum(self):
        rec = self.record
        if rec[0] != ":":
            return False        # if initial char is not a colon, cannot be a good record
        rec = rec[1:]
        if len(rec) % 2:
            return False        # since we do not have an even number of hex digits
                                #  there cannot be a good checksum
        sum = 0
        for i in range(0, len(rec), 2):
            sum += int (rec[i:i+2], 16)
        self.cksum = sum
        if sum % 0x100:
            return False       # bad checksum
        return True            # ok, everything checked out, so NOT a bad checksum

    def parseRecord (self):
        """Divide an Intel hex record into its various fields
        """
        rec = self.record
        self.length     = int (rec[1:3],   16)
        self.address    = int (rec[3:7],   16)
        self.type       = int (rec[7:9],   16)
        self.payloadfield = payloadfield  = rec[9:-2]
        if len(self.payloadfield) != (2 * self.length):
            # self.length is the byte length of the payload field
            #  the payload field contains 2 characters per byte
            print "length field %d does not match actual payload field %s in record:"
            print "'%s'" % self.record
            abortWithUsageMessage()
        if self.length % 2:
            # For the PIC, we expect 2 bytes per PIC word (thus 4 ascii hex characters)
            print "length field value %d is not an even number in record:"
            print "'%s'" % self.record
            abortWithUsageMessage()
        # The various checks look ok, so convert the payload field into an
        #  array of byte values in the correct order
        payload = [0xff] * self.length
        #print "DEBUG: (bytelength, payloadfield) = (%d, %s)" % (self.length, payloadfield)
        for w in range(0, (self.length / 2)):
            # process in pairs, where the first byte of the pair is the lsbyte
            fieldIndex = w*4
            #print "DEBUG:    (w, fieldIndex) = (%d, %d)" % (w, fieldIndex)
            lsbyte = int (payloadfield[fieldIndex : fieldIndex + 2], 16)
            msbyte = int (payloadfield[fieldIndex + 2 : fieldIndex + 4], 16)
            payload[w*2] = msbyte
            payload[(w*2)+1] = lsbyte
        self.payload = payload

def buildImage():
    """Set up an array of 0x4000 bytes, i.e. which is large enough to hold
       0x2000 PIC 16-bit words.  This is large enough for the largest PIC
       processor that we expect to use this downloader/bootloader combination
       with.  Prefill it with 0xFF bytes.  Use the hexFile to populate the
       image with the correct data.  Prior to transmitting the image to the
       PIC, it will be trimmed to the size implied by the hexFile, ignoring
       any configuration word (which can only be programmed by a hardware
       programmer anyway).  The first byte of the image corresponds to
       address zero in the PIC, although when we transmit to the PIC we will
       skip the first 32 words (64 bytes), as the bootloader expects to start
       with its second 32-word block.
    """
    global binaryImage
    imageSizeInBytes = 2 * 0x2000
    binaryImage = [0xff] * imageSizeInBytes
    print "initial byte length of binaryImage is %d" % len(binaryImage)
    gotEOF = False
    for i in hexFile:
        r = hexrec(i)
        if gotEOF:
            print "Warning: found type %d record *after* the EOF record -- continuing"
        if r.type == 1:
            gotEOF = True
        elif r.type == 4:
            extendedAddr = int(r.payloadfield, 16)
            if extendedAddr != 0:
                print "Warning: found type 4 (extended address) record with address = %x" % extendedAddr
                print "         continuing anyway as if the extended address were zero"
        elif r.type == 0:
            byteAddr = r.address    # this is a byte address (rather than a word address)
            for byte in r.payload:
                if byteAddr < imageSizeInBytes:
                    binaryImage[byteAddr] = byte
                    byteAddr += 1
        else:
            print "Warning: unexpected record with type %d -- ignoring" % r.type
    
def truncateImage():
    """We have allowed an image array that is typically much larger than
       needed by the PIC application, with the unused values set to the
       pad value of 0xff.  Truncate the trailing pad characters so we are
       left we the smallest image that contains the entire application.
       This way, we do not bother transmitting useless data to the bootloader.
    """
    global binaryImage
    indexOfLastNonPadByte = 0
    for i,v in enumerate(binaryImage):
        if v != 0xff:
            indexOfLastNonPadByte = i
    # However, we must not end the image in the middle of a word, even
    #  if the 2nd byte of the word happens to equal the pad byte.
    #  Also, we must end on a 4-word boundary, as the actual flash
    #  programming occurs in 4-word blocks.  So, we increment indexOfLastNonPadByte
    #  until its value mod 8 (i.e. word mod 4) is zero.  At that
    #  point indexOfLastNonPadByte is pointing to the first byte of the 4-word block
    #  just past the block we wish to keep.
    while (indexOfLastNonPadByte % 8):
        indexOfLastNonPadByte += 1
    binaryImage = binaryImage[:indexOfLastNonPadByte]
    print "byte length of binaryImage after truncating is %d" % len(binaryImage)
    # ** here we could test that the image is not too large for the target
    # **  processor's memory.  E.g. if downloading to the PIC16F819, whose
    # **  bootloader code in high-memory is org'd at about 0x0749 (a word address),
    # **  it would be bad to send an image large enough to overwrite the bootloader
    # **  code.

def serout (intval):
    """Convert the integer value to a character and transmit it via the
       serial port, then delay the appropriate delay.
    """
    ch = chr(intval)
    serialPort.write(ch)
    time.sleep (BYTEDELAY)

def seroutwait (intval):
    """Like serout() but wait for the character to be echo'd back
    """
    ch = chr(intval)
    serialPort.write(ch)
    while serialPort.read() != ch:
        pass

def transmitImage():
    """Send a few seconds worth of SYNC bytes to the PIC, to give time
       to power up the target PIC, then send the command
       ("P") to start programming, then send the count of words (as 2 bytes,
       msbyte first).  Then send the actual bytes.
       Wait an appropriate delay after sending the 2 bytes for the word
       that *begins* a 32-word block and also after the 2 bytes for the
       word that *ends* a 4-word block.  Allow an extra millisecond
       between every byte (i.e. a 2 ms sleep, the first for the approximately
       1 ms the serial port will spend transmitting the byte and the
       second for the delay between bytes).  For the 16F819, the 32-word
       and 4-word delays should be at least 2 ms.  For the 16F877, the
       4-word delay should be at least 4 ms (the 16F877 does not need a
       32-word delay).
    """
    print "Send sync bytes for a few seconds to give you time to turn on the PIC"
    #syncbytecount = 1250
    syncbytecount = 250
    for i in xrange(syncbytecount):
        serout (SYNC)
    print "Send 'P'"
    serout ( ord("P"))    # command to program the flash

    print "Send word count"
    wordcount = (len(binaryImage) / 2) - 32    # we do not transmit first 32-word block
    serout (wordcount // 0x100)   # send msbyte of word count
    serout (wordcount  % 0x100)   # send lsbyte of word count

    print "Send %d words" % wordcount
    for i in xrange(64, len(binaryImage), 2):
        # skip the first 32 words (i.e. the first 64 bytes) so that we begin
        # sending with the data for the 2nd 32-word block
        serout (binaryImage[i])     # send msbyte of word
        serout (binaryImage[i+1])   # send lsbyte of word
        # i is the byte address of the word just sent.  Divide by 2 to get word address.
        wordaddr = i // 2
        if (wordaddr % 32) == 0:
            # we just sent the first word for a new 32-word block, so give the
            # PIC time to do a block erase
            time.sleep (BLOCKDELAY)
        elif (wordaddr % 4) == 3:
            # we just sent the 3rd word of a 4-word block, so give the PIC
            #  time to do the actual programming
            time.sleep (BLOCKDELAY)
    print "Finished"

def dump (seq):
    """print a formatted dump in hexadecimal of the values in seq"""
    for i,v in enumerate(seq):
        if (i % 16) == 0:
            print                  # finish previous line
            print "%04x  " % i,      # print the address
        elif (i % 8) == 0:
            print "  ",             # gap between left 8 bytes and right 8 bytes
        print "%02x " % v,          # print the value in hexadecimal
    print
    
def picdump (seq):
    """Print a formatted dump in hexadecimal of the values in seq as in
       dump() but print two bytes at a time (word values) and show addresses
       as word addresses.  This should match the PIC assembler listing.
    """
    for i,v in enumerate(seq):
        if (i % 16) == 0:
            print                    # finish previous line
            print "%04x  " % (i/2),  # print the word address
        elif (i % 8) == 0:
            print "  ",             # gap between left 4 words and right 4 words
        elif (i % 2) == 0:
            print " ",              # print the gap between words
        print "%02x" % v,           # print the value in hexadecimal
    print

EX = 0
def echoIncoming():
    """To be run in a background thread.  Echo each serial input character.
       Start it somewhere with:
             t = threading.Thread(None, echoIncoming)
             t.start()
    """
    while EX == 0:
        c = serialPort.read()
        if c:
            # later we could handle crlf and non-printable characters
            sys.stdout.write(c)
            sys.stdout.flush()

def startEchoing():
    t = threading.Thread(None, echoIncoming)
    t.start()

def stop ():
    global EX
    EX = 1
    
    

# sample Intel hex records for testing
s1 = ":020000040000FA\n"
s2 = ":020000002428B2\n"
s3 = ":00000001FF\n"


if __name__ == '__main__':
    parseCommandLine()
    buildImage()
    truncateImage()
    transmitImage()
    hexFile.close()
    # comment out one or the other of the next two lines
    #startEchoing()
    serialPort.close()
    
# Local Variables:
# py-master-file: nil
# End:
